/*
 *  Copyright (C) 2002 Jorn Baayen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "pdm-dialog.h"
#include "galeon-shell.h"
#include "galeon-embed-shell.h"
#include "gul-gui.h"
#include "gul-ellipsizing-label.h"

#include <gtk/gtktreeview.h>
#include <gtk/gtkliststore.h>
#include <gtk/gtkcellrenderertext.h>
#include <bonobo/bonobo-i18n.h>

typedef struct PdmActionInfo PdmActionInfo;

typedef void (* PDM_add) (PdmActionInfo *info, gpointer data);
typedef void (* PDM_remove) (PdmActionInfo *info, GList *data);
typedef void (* PDM_free) (PdmActionInfo *info, GList *data);

static void pdm_dialog_class_init (PdmDialogClass *klass);
static void pdm_dialog_init (PdmDialog *dialog);
static void pdm_dialog_finalize (GObject *object);

/* Glade callbacks */
void
pdm_dialog_close_button_clicked_cb (GtkWidget *button,
			            PdmDialog *dialog);
void
pdm_dialog_cookies_properties_button_clicked_cb (GtkWidget *button,
						 PdmDialog *dialog);
void
pdm_dialog_cookies_treeview_selection_changed_cb (GtkTreeSelection *selection,
                                          	  PdmDialog *dialog);
void
pdm_dialog_passwords_treeview_selection_changed_cb (GtkTreeSelection *selection,
                                          	    PdmDialog *dialog);

static GObjectClass *parent_class = NULL;

struct PdmActionInfo
{
	PDM_add add;
	PDM_remove remove;
	PDM_free free;
	GtkWidget *treeview;
	GList *list;
	int remove_id;
	int data_col;
	PdmDialog *dialog;
};

struct PdmDialogPrivate
{
	GtkTreeModel *model;
	PdmActionInfo *cookies;
	PdmActionInfo *passwords;
	PdmActionInfo *password_sites;
	PdmActionInfo *cookie_sites;
	PdmActionInfo *image_sites;
};

enum
{
	PROP_DIALOG,

	PROP_COOKIES_TREEVIEW,
	PROP_COOKIES_REMOVE,
	PROP_COOKIES_PROPERTIES,

	PROP_PASSWORDS_TREEVIEW,
	PROP_PASSWORDS_REMOVE,

	PROP_COOKIE_SITES_TREEVIEW,
	PROP_COOKIE_SITES_REMOVE,

	PROP_IMAGE_SITES_TREEVIEW,
	PROP_IMAGE_SITES_REMOVE,

	PROP_PASSWORD_SITES_TREEVIEW,
	PROP_PASSWORD_SITES_REMOVE,
};

enum
{
	COL_COOKIES_HOST,
	COL_COOKIES_NAME,
	COL_COOKIES_DATA
};

enum
{
	COL_PASSWORDS_HOST,
	COL_PASSWORDS_USER,
	COL_PASSWORDS_DATA
};

enum
{
	COL_COOKIE_SITES_DOMAIN,
	COL_COOKIE_SITES_STATUS,
	COL_COOKIE_SITES_DATA
};

enum
{
	COL_IMAGE_SITES_DOMAIN,
	COL_IMAGE_SITES_STATUS,
	COL_IMAGE_SITES_DATA
};

enum
{
	COL_PASSWORD_SITES_HOST,
	COL_PASSWORD_SITES_STATUS,
	COL_PASSWORD_SITES_DATA
};

static const
GaleonDialogProperty properties [] =
{
	{ PROP_DIALOG, "pdm_dialog", NULL, PT_NORMAL, NULL },

	{ PROP_COOKIES_TREEVIEW, "cookies_treeview", NULL, PT_NORMAL, NULL },
	{ PROP_COOKIES_REMOVE, "cookies_remove_button", NULL, PT_NORMAL, NULL },
	{ PROP_COOKIES_PROPERTIES, "cookies_properties_button", NULL, PT_NORMAL, NULL },

	{ PROP_PASSWORDS_TREEVIEW, "passwords_treeview", NULL, PT_NORMAL, NULL },
	{ PROP_PASSWORDS_REMOVE, "passwords_remove_button", NULL, PT_NORMAL, NULL },

	{ PROP_COOKIE_SITES_TREEVIEW, "cookie_sites_treeview", NULL, PT_NORMAL, NULL },
	{ PROP_COOKIE_SITES_REMOVE, "cookie_sites_remove_button", NULL, PT_NORMAL, NULL },

	{ PROP_IMAGE_SITES_TREEVIEW, "image_sites_treeview", NULL, PT_NORMAL, NULL },
	{ PROP_IMAGE_SITES_REMOVE, "image_sites_remove_button", NULL, PT_NORMAL, NULL },

	{ PROP_PASSWORD_SITES_TREEVIEW, "password_sites_treeview", NULL, PT_NORMAL, NULL },
	{ PROP_PASSWORD_SITES_REMOVE, "password_sites_remove_button", NULL, PT_NORMAL, NULL },

	{ -1, NULL, NULL }
};

GType 
pdm_dialog_get_type (void)
{
        static GType pdm_dialog_type = 0;

        if (pdm_dialog_type == 0)
        {
                static const GTypeInfo our_info =
                {
                        sizeof (PdmDialogClass),
                        NULL, /* base_init */
                        NULL, /* base_finalize */
                        (GClassInitFunc) pdm_dialog_class_init,
                        NULL,
                        NULL, /* class_data */
                        sizeof (PdmDialog),
                        0, /* n_preallocs */
                        (GInstanceInitFunc) pdm_dialog_init
                };

                pdm_dialog_type = g_type_register_static (GALEON_DIALOG_TYPE,
						              "PdmDialog",
						              &our_info, 0);
        }

        return pdm_dialog_type;

}

static void
pdm_dialog_class_init (PdmDialogClass *klass)
{
        GObjectClass *object_class = G_OBJECT_CLASS (klass);

        parent_class = g_type_class_peek_parent (klass);

        object_class->finalize = pdm_dialog_finalize;
}

static void
cookies_treeview_selection_changed_cb (GtkTreeSelection *selection,
                                       PdmDialog *dialog)
{
	GtkWidget *widget;
	GList *l;
	GaleonDialog *d = GALEON_DIALOG(dialog);
	gboolean can_properties;
	
	l = gul_gui_treeview_get_selection_refs 
		(GTK_TREE_VIEW(dialog->priv->cookies->treeview));

	can_properties = (l != NULL && l->next == NULL);
	gul_gui_treeview_free_selection_refs (l);
	
	widget = galeon_dialog_get_control (d, PROP_COOKIES_PROPERTIES);
	gtk_widget_set_sensitive (widget, can_properties);
}

static void
action_treeview_selection_changed_cb (GtkTreeSelection *selection,
                                      PdmActionInfo *action)
{
	GtkWidget *widget;
	GList *l;
	GaleonDialog *d = GALEON_DIALOG(action->dialog);
	gboolean has_selection;
	
	l = gul_gui_treeview_get_selection_refs 
		(GTK_TREE_VIEW(action->treeview));

	has_selection = l != NULL;
	gul_gui_treeview_free_selection_refs (l);

	widget = galeon_dialog_get_control (d, action->remove_id);
	gtk_widget_set_sensitive (widget, has_selection);
}

static GtkWidget *
setup_passwords_treeview (PdmDialog *dialog)
{

	GtkTreeView *treeview;
        GtkListStore *liststore;
        GtkCellRenderer *renderer;
        GtkTreeViewColumn *column;
	GtkTreeSelection *selection;

	treeview = GTK_TREE_VIEW(galeon_dialog_get_control 
				 (GALEON_DIALOG(dialog),
				 PROP_PASSWORDS_TREEVIEW));
	
        /* set tree model */
        liststore = gtk_list_store_new (3,
                                        G_TYPE_STRING,
                                        G_TYPE_STRING,
                                        G_TYPE_POINTER);
        gtk_tree_view_set_model (treeview, GTK_TREE_MODEL(liststore));
        gtk_tree_view_set_headers_visible (treeview, TRUE);
	selection = gtk_tree_view_get_selection (treeview);
	gtk_tree_selection_set_mode (selection,
				     GTK_SELECTION_MULTIPLE);
	
        renderer = gtk_cell_renderer_text_new ();

        gtk_tree_view_insert_column_with_attributes (treeview,
                                                     COL_PASSWORDS_HOST, 
						     _("Host"),
                                                     renderer,
                                                     "text", COL_PASSWORDS_HOST,
                                                     NULL);
        column = gtk_tree_view_get_column (treeview, COL_PASSWORDS_HOST);
        gtk_tree_view_column_set_resizable (column, TRUE);
        gtk_tree_view_column_set_reorderable (column, TRUE);
        gtk_tree_view_column_set_sizing (column, GTK_TREE_VIEW_COLUMN_AUTOSIZE);
        gtk_tree_view_column_set_sort_column_id (column, COL_PASSWORDS_HOST);

        gtk_tree_view_insert_column_with_attributes (treeview,
                                                     COL_PASSWORDS_USER, 
						     _("User Name"),
                                                     renderer,
                                                     "text", COL_PASSWORDS_USER,
                                                     NULL);
        column = gtk_tree_view_get_column (treeview, COL_PASSWORDS_USER);
        gtk_tree_view_column_set_resizable (column, TRUE);
        gtk_tree_view_column_set_reorderable (column, TRUE);
        gtk_tree_view_column_set_sizing (column, GTK_TREE_VIEW_COLUMN_AUTOSIZE);
        gtk_tree_view_column_set_sort_column_id (column, COL_PASSWORDS_USER);

	return GTK_WIDGET (treeview);
}

static GtkWidget *
setup_password_sites_treeview (PdmDialog *dialog)
{

	GtkTreeView *treeview;
        GtkListStore *liststore;
        GtkCellRenderer *renderer;
        GtkTreeViewColumn *column;
	GtkTreeSelection *selection;

	treeview = GTK_TREE_VIEW(galeon_dialog_get_control 
				 (GALEON_DIALOG(dialog),
				 PROP_PASSWORD_SITES_TREEVIEW));
	
        /* set tree model */
        liststore = gtk_list_store_new(3,
                                       G_TYPE_STRING,
                                       G_TYPE_STRING,
                                       G_TYPE_POINTER);
        gtk_tree_view_set_model(treeview, GTK_TREE_MODEL(liststore));
        gtk_tree_view_set_headers_visible(treeview, TRUE);
	selection = gtk_tree_view_get_selection(treeview);
	gtk_tree_selection_set_mode(selection,
				    GTK_SELECTION_MULTIPLE);
	
        renderer = gtk_cell_renderer_text_new();

        gtk_tree_view_insert_column_with_attributes (treeview,
                                                     COL_PASSWORD_SITES_HOST, 
						     _("Host"),
                                                     renderer,
                                                     "text", COL_PASSWORD_SITES_HOST,
                                                     NULL);
        column = gtk_tree_view_get_column(treeview, COL_PASSWORD_SITES_HOST);
        gtk_tree_view_column_set_resizable(column, TRUE);
        gtk_tree_view_column_set_reorderable(column, TRUE);
        gtk_tree_view_column_set_sizing(column, GTK_TREE_VIEW_COLUMN_AUTOSIZE);
        gtk_tree_view_column_set_sort_column_id(column, COL_PASSWORD_SITES_HOST);

	gtk_tree_view_insert_column_with_attributes(treeview,
                                                    COL_PASSWORD_SITES_STATUS, 
						    _("Status"),
                                                    renderer,
                                                    "text", COL_PASSWORD_SITES_STATUS,
                                                    NULL);
        column = gtk_tree_view_get_column(treeview, COL_PASSWORD_SITES_STATUS);
        gtk_tree_view_column_set_resizable(column, TRUE);
        gtk_tree_view_column_set_reorderable(column, TRUE);
        gtk_tree_view_column_set_sizing(column, GTK_TREE_VIEW_COLUMN_AUTOSIZE);
        gtk_tree_view_column_set_sort_column_id(column, COL_PASSWORD_SITES_STATUS);

	return GTK_WIDGET(treeview);
}

static GtkWidget *
setup_cookies_treeview (PdmDialog *dialog)
{
	GtkTreeView *treeview;
        GtkListStore *liststore;
        GtkCellRenderer *renderer;
        GtkTreeViewColumn *column;
	GtkTreeSelection *selection;
	
	treeview = GTK_TREE_VIEW (galeon_dialog_get_control 
				  (GALEON_DIALOG(dialog),
				  PROP_COOKIES_TREEVIEW));

	
	
        /* set tree model */
        liststore = gtk_list_store_new (3,
                                        G_TYPE_STRING,
                                        G_TYPE_STRING,
                                        G_TYPE_POINTER);
        gtk_tree_view_set_model (treeview, GTK_TREE_MODEL(liststore));
        gtk_tree_view_set_headers_visible (treeview, TRUE);
	selection = gtk_tree_view_get_selection (treeview);
	gtk_tree_selection_set_mode (selection,
				     GTK_SELECTION_MULTIPLE);

	g_signal_connect (selection, "changed",
			  G_CALLBACK(cookies_treeview_selection_changed_cb),
			  dialog);
	
	renderer = gtk_cell_renderer_text_new ();

	gtk_tree_view_insert_column_with_attributes (treeview,
                                                     COL_COOKIES_HOST, 
						     _("Domain"),
                                                     renderer,
                                                     "text", COL_COOKIES_HOST,
                                                     NULL);
        column = gtk_tree_view_get_column (treeview, COL_COOKIES_HOST);
        gtk_tree_view_column_set_resizable (column, TRUE);
        gtk_tree_view_column_set_reorderable (column, TRUE);
        gtk_tree_view_column_set_sizing (column, GTK_TREE_VIEW_COLUMN_AUTOSIZE);
        gtk_tree_view_column_set_sort_column_id (column, COL_COOKIES_HOST);
	
        gtk_tree_view_insert_column_with_attributes (treeview,
                                                     COL_COOKIES_NAME, 
						     _("Name"),
                                                     renderer,
                                                     "text", COL_COOKIES_NAME,
                                                     NULL);
        column = gtk_tree_view_get_column (treeview, COL_COOKIES_NAME);
        gtk_tree_view_column_set_resizable (column, TRUE);
        gtk_tree_view_column_set_reorderable (column, TRUE);
        gtk_tree_view_column_set_sizing (column, GTK_TREE_VIEW_COLUMN_AUTOSIZE);
        gtk_tree_view_column_set_sort_column_id (column, COL_COOKIES_NAME);

	return GTK_WIDGET(treeview);
}

static GtkWidget *
setup_cookie_sites_treeview (PdmDialog *dialog)
{
	GtkTreeView *treeview;
        GtkListStore *liststore;
        GtkCellRenderer *renderer;
        GtkTreeViewColumn *column;
	GtkTreeSelection *selection;
	
	treeview = GTK_TREE_VIEW (galeon_dialog_get_control 
				  (GALEON_DIALOG(dialog),
				  PROP_COOKIE_SITES_TREEVIEW));

	
	
        /* set tree model */
        liststore = gtk_list_store_new (3,
                                        G_TYPE_STRING,
                                        G_TYPE_STRING,
                                        G_TYPE_POINTER);
        gtk_tree_view_set_model (treeview, GTK_TREE_MODEL(liststore));
        gtk_tree_view_set_headers_visible (treeview, TRUE);
	selection = gtk_tree_view_get_selection (treeview);
	gtk_tree_selection_set_mode (selection,
				     GTK_SELECTION_MULTIPLE);

	renderer = gtk_cell_renderer_text_new ();

	gtk_tree_view_insert_column_with_attributes (treeview,
                                                     COL_COOKIE_SITES_DOMAIN, 
						     _("Domain"),
                                                     renderer,
                                                     "text", COL_COOKIE_SITES_DOMAIN,
                                                     NULL);
        column = gtk_tree_view_get_column (treeview, COL_COOKIE_SITES_DOMAIN);
        gtk_tree_view_column_set_resizable (column, TRUE);
        gtk_tree_view_column_set_reorderable (column, TRUE);
        gtk_tree_view_column_set_sizing (column, GTK_TREE_VIEW_COLUMN_AUTOSIZE);
        gtk_tree_view_column_set_sort_column_id (column, COL_COOKIE_SITES_DOMAIN);
	
        gtk_tree_view_insert_column_with_attributes (treeview,
                                                     COL_COOKIE_SITES_STATUS, 
						     _("Status"),
                                                     renderer,
                                                     "text", COL_COOKIE_SITES_STATUS,
                                                     NULL);
        column = gtk_tree_view_get_column (treeview, COL_COOKIE_SITES_STATUS);
        gtk_tree_view_column_set_resizable (column, TRUE);
        gtk_tree_view_column_set_reorderable (column, TRUE);
        gtk_tree_view_column_set_sizing (column, GTK_TREE_VIEW_COLUMN_AUTOSIZE);
        gtk_tree_view_column_set_sort_column_id (column, COL_COOKIE_SITES_STATUS);

	return GTK_WIDGET(treeview);
}

static GtkWidget *
setup_image_sites_treeview (PdmDialog *dialog)
{
	GtkTreeView *treeview;
        GtkListStore *liststore;
        GtkCellRenderer *renderer;
        GtkTreeViewColumn *column;
	GtkTreeSelection *selection;
	
	treeview = GTK_TREE_VIEW(galeon_dialog_get_control 
				 (GALEON_DIALOG(dialog),
				 PROP_IMAGE_SITES_TREEVIEW));

	
	
        /* set tree model */
        liststore = gtk_list_store_new(3,
                                       G_TYPE_STRING,
                                       G_TYPE_STRING,
                                       G_TYPE_POINTER);
        gtk_tree_view_set_model(treeview, GTK_TREE_MODEL(liststore));
        gtk_tree_view_set_headers_visible(treeview, TRUE);
	selection = gtk_tree_view_get_selection(treeview);
	gtk_tree_selection_set_mode(selection,
				    GTK_SELECTION_MULTIPLE);

	renderer = gtk_cell_renderer_text_new();

	gtk_tree_view_insert_column_with_attributes(treeview,
                                                    COL_IMAGE_SITES_DOMAIN, 
						    _("Domain"),
                                                    renderer,
                                                    "text", COL_IMAGE_SITES_DOMAIN,
                                                    NULL);
        column = gtk_tree_view_get_column(treeview, COL_IMAGE_SITES_DOMAIN);
        gtk_tree_view_column_set_resizable(column, TRUE);
        gtk_tree_view_column_set_reorderable(column, TRUE);
        gtk_tree_view_column_set_sizing(column, GTK_TREE_VIEW_COLUMN_AUTOSIZE);
        gtk_tree_view_column_set_sort_column_id(column, COL_IMAGE_SITES_DOMAIN);
	
        gtk_tree_view_insert_column_with_attributes(treeview,
                                                    COL_IMAGE_SITES_STATUS, 
						    _("Status"),
                                                    renderer,
                                                    "text", COL_IMAGE_SITES_STATUS,
                                                    NULL);
        column = gtk_tree_view_get_column(treeview, COL_IMAGE_SITES_STATUS);
        gtk_tree_view_column_set_resizable(column, TRUE);
        gtk_tree_view_column_set_reorderable(column, TRUE);
        gtk_tree_view_column_set_sizing(column, GTK_TREE_VIEW_COLUMN_AUTOSIZE);
        gtk_tree_view_column_set_sort_column_id(column, COL_IMAGE_SITES_STATUS);

	return GTK_WIDGET(treeview);
}

static void
pdm_dialog_remove_button_clicked_cb (GtkWidget *button,
				     PdmActionInfo *action)
{
	GList *refs, *l;
	GList *remove_list = NULL;
	GtkTreeModel *model;

	model = gtk_tree_view_get_model 
		(GTK_TREE_VIEW(action->treeview));
	refs = gul_gui_treeview_get_selection_refs 
		(GTK_TREE_VIEW(action->treeview));

	for (l = refs; l != NULL; l = l->next)
	{
		GtkTreeIter iter;
		gpointer data;
		GtkTreePath *path;
		GValue val = {0, };

		path = gtk_tree_row_reference_get_path 
			((GtkTreeRowReference *)l->data); 
        	gtk_tree_model_get_iter 
			(model, &iter, path);
		gtk_tree_model_get_value 
			(model, &iter, action->data_col, &val);
		data = g_value_get_pointer (&val);

		gtk_list_store_remove (GTK_LIST_STORE(model),
				       &iter);
		
		action->list = g_list_remove (action->list, data);
		remove_list = g_list_append (remove_list, data);
	}

	gul_gui_treeview_free_selection_refs (refs);

	if (remove_list)
	{
		action->remove (action, remove_list);
		action->free (action, remove_list);
	}
}

static void
setup_action (PdmActionInfo *action)
{
	GList *l;
	GtkWidget *widget;
	GtkTreeSelection *selection;
	
	for (l = action->list; l != NULL; l = l->next)
	{
		action->add (action, l->data);
	}

	widget = galeon_dialog_get_control (GALEON_DIALOG(action->dialog), 
					    action->remove_id);
	g_signal_connect (widget, "clicked",
			  G_CALLBACK(pdm_dialog_remove_button_clicked_cb),
			  action);
	
	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW(action->treeview));
	g_signal_connect (selection, "changed",
			  G_CALLBACK(action_treeview_selection_changed_cb),
			  action);
}

static void
pdm_dialog_cookie_add (PdmActionInfo *info, 
		       gpointer cookie)
{
	GtkListStore *store;
	GtkTreeIter iter;
	CookieInfo *cinfo = (CookieInfo *)cookie;
	
	store = GTK_LIST_STORE(gtk_tree_view_get_model 
			       (GTK_TREE_VIEW(info->treeview)));

	gtk_list_store_append (store, &iter);
        gtk_list_store_set (store,
                            &iter,
			    COL_COOKIES_HOST, cinfo->domain,
                            COL_COOKIES_NAME, cinfo->name,
			    COL_COOKIES_DATA, cinfo,
                            -1);
}

static void
pdm_dialog_password_add (PdmActionInfo *info, 
			 gpointer password)
{
	GtkListStore *store;
	GtkTreeIter iter;
	PasswordInfo *pinfo = (PasswordInfo *)password;

	store = GTK_LIST_STORE(gtk_tree_view_get_model 
			       (GTK_TREE_VIEW(info->treeview)));

	gtk_list_store_append (store, &iter);
        gtk_list_store_set (store,
                            &iter,
                            COL_PASSWORDS_HOST, pinfo->host,
                            COL_PASSWORDS_USER, pinfo->username,
			    COL_PASSWORDS_DATA, pinfo,
                            -1);
}

static void
pdm_dialog_password_site_add(PdmActionInfo *info, 
			     gpointer password_site)
{
	GtkListStore *store;
	GtkTreeIter iter;
	PasswordInfo *pinfo = (PasswordInfo *)password_site;

	store = GTK_LIST_STORE(gtk_tree_view_get_model 
			       (GTK_TREE_VIEW(info->treeview)));

	gtk_list_store_append(store, &iter);
        gtk_list_store_set(store,
                           &iter,
                           COL_PASSWORD_SITES_HOST, pinfo->host,
                           COL_PASSWORD_SITES_STATUS, _("Cannot save passwords"),
			   COL_PASSWORD_SITES_DATA, pinfo,
                           -1);
}

static void
pdm_dialog_cookie_site_add (PdmActionInfo *info, 
		            gpointer cookie_site)
{
	GtkListStore *store;
	GtkTreeIter iter;
	PermissionInfo *pinfo = (PermissionInfo *)cookie_site;
	
	store = GTK_LIST_STORE(gtk_tree_view_get_model 
			       (GTK_TREE_VIEW(info->treeview)));

	gtk_list_store_append (store, &iter);
        gtk_list_store_set (store,
                            &iter,
			    COL_COOKIE_SITES_DOMAIN, pinfo->domain,
                            COL_COOKIE_SITES_STATUS, pinfo->status 
			                             ? _("Can set cookies") 
						     : _("Cannot set cookies"),
			    COL_COOKIE_SITES_DATA,   pinfo,
                            -1);
}

static void
pdm_dialog_image_site_add (PdmActionInfo *info, 
		            gpointer image_site)
{
	GtkListStore *store;
	GtkTreeIter iter;
	PermissionInfo *pinfo = (PermissionInfo *)image_site;
	
	store = GTK_LIST_STORE(gtk_tree_view_get_model 
			       (GTK_TREE_VIEW(info->treeview)));

	gtk_list_store_append(store, &iter);
        gtk_list_store_set(store,
                           &iter,
			   COL_IMAGE_SITES_DOMAIN, pinfo->domain,
                           COL_IMAGE_SITES_STATUS, pinfo->status 
			                           ? _("Can show images") 
						   : _("Cannot show images"),
			   COL_IMAGE_SITES_DATA,   pinfo,
                           -1);
}

static void
pdm_dialog_cookie_remove (PdmActionInfo *info,
			  GList *data)
{	
	GaleonEmbedShell *shell;
	shell = galeon_shell_get_embed_shell (galeon_shell);
	galeon_embed_shell_remove_cookies (shell, data);
}

static void
pdm_dialog_password_remove (PdmActionInfo *info,
			    GList *data)
{
	GaleonEmbedShell *shell;
	shell = galeon_shell_get_embed_shell (galeon_shell);
	galeon_embed_shell_remove_passwords (shell, data, PASSWORD_PASSWORD);
}

static void
pdm_dialog_password_site_remove(PdmActionInfo *info,
				GList *data)
{
	GaleonEmbedShell *shell;
	shell = galeon_shell_get_embed_shell(galeon_shell);
	galeon_embed_shell_remove_passwords(shell, data, PASSWORD_REJECT);
}

static void
pdm_dialog_cookie_site_remove (PdmActionInfo *info,
			       GList *data)
{
	GaleonEmbedShell *shell;
	shell = galeon_shell_get_embed_shell (galeon_shell);
	galeon_embed_shell_remove_permissions (shell, COOKIES_PERMISSION, data);
}

static void
pdm_dialog_image_site_remove(PdmActionInfo *info,
			     GList *data)
{
	GaleonEmbedShell *shell;
	shell = galeon_shell_get_embed_shell(galeon_shell);
	galeon_embed_shell_remove_permissions(shell, IMAGES_PERMISSION, data);
}

static void
pdm_dialog_cookies_free (PdmActionInfo *info,
			 GList *data)
{	
	GaleonEmbedShell *shell;
	GList *l;
		
	shell = galeon_shell_get_embed_shell (galeon_shell);
	l = data ? data : info->list;
	galeon_embed_shell_free_cookies (shell, l);
}

static void
pdm_dialog_passwords_free (PdmActionInfo *info,
			   GList *data)
{
	GaleonEmbedShell *shell;
	GList *l;
	
	shell = galeon_shell_get_embed_shell (galeon_shell);
	l = data ? data : info->list;
	galeon_embed_shell_free_passwords (shell, l);
}

static void
pdm_dialog_site_permissions_free(PdmActionInfo *info,
				 GList *data)
{
	GaleonEmbedShell *shell;
	GList *l;
	
	shell = galeon_shell_get_embed_shell (galeon_shell);
	l = data ? data : info->list;
	galeon_embed_shell_free_permissions (shell, l);
}

/* Group all Properties and Remove buttons in the same size group to avoid the
 * little jerk you get otherwise when switching pages because one set of
 * buttons is wider than another. */
static void
group_button_allocations (GaleonDialog *dialog)
{
	const gint props[] =
	{
		PROP_COOKIES_REMOVE,
		PROP_COOKIES_PROPERTIES,
		PROP_PASSWORDS_REMOVE,
		PROP_COOKIE_SITES_REMOVE,
		PROP_IMAGE_SITES_REMOVE,
		PROP_PASSWORD_SITES_REMOVE,
	};
	GtkSizeGroup *size_group;
	guint i;

	size_group = gtk_size_group_new (GTK_SIZE_GROUP_HORIZONTAL);

	for (i = 0; i < G_N_ELEMENTS (props); ++i)
	{
		GtkWidget *w;
		w = galeon_dialog_get_control (dialog, props[i]);
		gtk_size_group_add_widget (size_group,  w);
	}
}

static void
pdm_dialog_init (PdmDialog *dialog)
{	
	GaleonEmbedShell *shell;
	PdmActionInfo *cookies;
	PdmActionInfo *passwords;
	PdmActionInfo *cookie_sites;
	PdmActionInfo *image_sites;
	PdmActionInfo *password_sites;
	GtkWidget *cookies_tv;
	GtkWidget *passwords_tv;
	GtkWidget *cookie_sites_tv;
	GtkWidget *image_sites_tv;
	GtkWidget *password_sites_tv;
	
	shell = galeon_shell_get_embed_shell (galeon_shell);
	
	dialog->priv = g_new0 (PdmDialogPrivate, 1);
	dialog->priv->cookies = NULL;
	dialog->priv->passwords = NULL;
	
	galeon_dialog_construct (GALEON_DIALOG(dialog),
                                 properties,
                                 "galeon.glade",
                                 "pdm_dialog");

	group_button_allocations (GALEON_DIALOG (dialog));

	cookies_tv = setup_cookies_treeview (dialog);
	passwords_tv = setup_passwords_treeview (dialog);
	cookie_sites_tv = setup_cookie_sites_treeview (dialog);
	image_sites_tv = setup_image_sites_treeview(dialog);
	password_sites_tv = setup_password_sites_treeview(dialog);

	cookies = g_new0 (PdmActionInfo, 1);
	galeon_embed_shell_list_cookies (shell, &cookies->list);
	cookies->dialog = dialog;
	cookies->remove_id = PROP_COOKIES_REMOVE;
	cookies->add = pdm_dialog_cookie_add;
	cookies->remove = pdm_dialog_cookie_remove;
	cookies->free = pdm_dialog_cookies_free;
	cookies->treeview = cookies_tv;
	cookies->data_col = COL_COOKIES_DATA;
	setup_action (cookies);

	passwords = g_new0 (PdmActionInfo, 1);
	galeon_embed_shell_list_passwords (shell, PASSWORD_PASSWORD,
					   &passwords->list);
	passwords->dialog = dialog;
	passwords->remove_id = PROP_PASSWORDS_REMOVE;
	passwords->add = pdm_dialog_password_add;
	passwords->remove = pdm_dialog_password_remove;
	passwords->free = pdm_dialog_passwords_free;
	passwords->treeview = passwords_tv;
	passwords->data_col = COL_PASSWORDS_DATA;
	setup_action (passwords);

	password_sites = g_new0(PdmActionInfo, 1);
	galeon_embed_shell_list_passwords(shell, PASSWORD_REJECT,
					  &password_sites->list);
	password_sites->dialog = dialog;
	password_sites->remove_id = PROP_PASSWORD_SITES_REMOVE;
	password_sites->add = pdm_dialog_password_site_add;
	password_sites->remove = pdm_dialog_password_site_remove;
	password_sites->free = pdm_dialog_passwords_free;
	password_sites->treeview = password_sites_tv;
	password_sites->data_col = COL_PASSWORD_SITES_DATA;
	setup_action(password_sites);

	cookie_sites = g_new0 (PdmActionInfo, 1);
	galeon_embed_shell_list_permissions (shell, COOKIES_PERMISSION,
					     &cookie_sites->list);
	cookie_sites->dialog	= dialog;
	cookie_sites->remove_id = PROP_COOKIE_SITES_REMOVE;
	cookie_sites->add	= pdm_dialog_cookie_site_add;
	cookie_sites->remove	= pdm_dialog_cookie_site_remove;
	cookie_sites->free	= pdm_dialog_site_permissions_free;
	cookie_sites->treeview	= cookie_sites_tv;
	cookie_sites->data_col	= COL_COOKIE_SITES_DATA;
	setup_action (cookie_sites);

	image_sites = g_new0(PdmActionInfo, 1);
	galeon_embed_shell_list_permissions(shell, IMAGES_PERMISSION,
					    &image_sites->list);
	image_sites->dialog	= dialog;
	image_sites->remove_id	= PROP_IMAGE_SITES_REMOVE;
	image_sites->add	= pdm_dialog_image_site_add;
	image_sites->remove	= pdm_dialog_image_site_remove;
	image_sites->free	= pdm_dialog_site_permissions_free;
	image_sites->treeview	= image_sites_tv;
	image_sites->data_col	= COL_IMAGE_SITES_DATA;
	setup_action(image_sites);

	dialog->priv->cookies = cookies;
	dialog->priv->passwords = passwords;
	dialog->priv->password_sites = password_sites;
	dialog->priv->cookie_sites = cookie_sites;
	dialog->priv->image_sites = image_sites;
}

static void
pdm_dialog_finalize (GObject *object)
{
	PdmDialog *dialog;

        g_return_if_fail (object != NULL);
        g_return_if_fail (IS_PDM_DIALOG (object));

	dialog = PDM_DIALOG (object);

        g_return_if_fail (dialog->priv != NULL);

	pdm_dialog_site_permissions_free (dialog->priv->cookie_sites, NULL);
	pdm_dialog_site_permissions_free (dialog->priv->image_sites, NULL);
	pdm_dialog_passwords_free (dialog->priv->passwords, NULL);
	pdm_dialog_passwords_free (dialog->priv->password_sites, NULL);
	pdm_dialog_cookies_free (dialog->priv->cookies, NULL);
	
        g_free (dialog->priv);

        G_OBJECT_CLASS (parent_class)->finalize (object);
}

GaleonDialog *
pdm_dialog_new (GtkWidget *window)
{
	PdmDialog *dialog;
	
	dialog = PDM_DIALOG (g_object_new (PDM_DIALOG_TYPE,
				     	   "ParentWindow", window,
				           NULL));

	return GALEON_DIALOG(dialog);
}

static void
show_cookies_properties (PdmDialog *dialog,
			 const CookieInfo *info)
{
	GtkWidget *table;
	GtkWidget *gdialog;
	GtkWidget *parent;
	GtkWidget *dialog_vbox;

	parent = galeon_dialog_get_control (GALEON_DIALOG(dialog),
					    PROP_DIALOG);
	
	gdialog = gtk_dialog_new_with_buttons 
		 (_("Cookie properties"), 
		  GTK_WINDOW(parent), 
		  GTK_DIALOG_MODAL, 
		  GTK_STOCK_CLOSE, 0, NULL);	
	gtk_dialog_set_has_separator (GTK_DIALOG(gdialog), FALSE);
	gtk_container_set_border_width (GTK_CONTAINER(gdialog), 6);

	table = cookie_info_table_new (info);
	dialog_vbox = GTK_DIALOG(gdialog)->vbox;
	gtk_box_pack_start (GTK_BOX(dialog_vbox), 
                            table,
                            FALSE, FALSE, 0);
	
	gtk_dialog_run (GTK_DIALOG(gdialog));

	gtk_widget_destroy (gdialog);
}

void
pdm_dialog_cookies_properties_button_clicked_cb (GtkWidget *button,
						 PdmDialog *dialog)
{
	GtkTreeModel *model;
	GValue val = {0, };
	GtkTreeIter iter;
	GtkTreePath *path;
	CookieInfo *info;
	GList *l;
	GtkWidget *treeview = dialog->priv->cookies->treeview;
	
	model = gtk_tree_view_get_model 
		(GTK_TREE_VIEW(treeview));
	l = gul_gui_treeview_get_selection_refs 
		(GTK_TREE_VIEW(treeview));

	path = gtk_tree_row_reference_get_path 
		((GtkTreeRowReference *)l->data); 
	gul_gui_treeview_free_selection_refs (l);

        gtk_tree_model_get_iter (model, &iter, path);
	gtk_tree_model_get_value 
		(model, &iter, COL_COOKIES_DATA, &val);
	info = (CookieInfo *)g_value_get_pointer (&val);

	show_cookies_properties (dialog, info);
}

void
pdm_dialog_close_button_clicked_cb (GtkWidget *button,
			            PdmDialog *dialog)
{
	g_object_unref (dialog);
}
